; #########################################################################
;
;                             ascii2qw
;
; Procedure for converting a signed null-terminated ascii string from memory
; to a 64-bit signed QWORD integer to memory.
;
; Limitations:    Up to 19 numerical digits are allowed
;                 but must not exceed +/-9223372036854775807
;                 The st(7) registers on the FPU will be trashed
;
; Return: EAX != 0 if ASCII string acceptable
;         EAX  = 0 if FPU unusable or string not acceptable
;
; Unacceptable strings are:
; - a null string
; - more than 18 numerical characters
; - characters other than numerical characters and one leading sign before
;   the terminating 0 (however, any number of leading spaces is allowed)
;
; Usage:  invoke ascii2qw, lpsz, lpqw
;
;   where:  
;   lpsz  = pointer to the null-terminated ASCII string in memory
;   lpqw  = pointer to a memory area to store the resulting QWORD
;
; Remark:
; It is the programmer's responsibility to check the error/success code
; in EAX before using the qword value at the supplied address.
;
;                       by Raymond Filiatreault
;                            February 2007
;                            MASM  syntax
;
; #########################################################################
;
;      .386                   ; minimum processor needed for 32 bit
;      .model flat, stdcall   ; FLAT memory model & STDCALL calling
;      option casemap :none   ; set code to case sensitive
;
; #########################################################################

ascii2qw PROTO :DWORD, :DWORD

.code

ascii2qw proc USES esi edi lpsz:DWORD, lpqw:DWORD

LOCAL bcdstring   :TBYTE

      mov   esi,lpsz          ;ESI = address of ascii string
      lea   edi,bcdstring
      xor   eax,eax
      stosd                   ;rezero the tbyte
      stosd
      stosw

;skip leading spaces first
   @@:
      cmp   byte ptr[esi]," "
      jnz   @F
      inc   esi
      jmp   @B
   @@:

;check if empty string
      cmp   byte ptr[esi],0
      jnz   @F
exit_error:
      xor   eax,eax
      ret

;check if leading sign
   @@:
      lea   edi,bcdstring
      cmp   byte ptr[esi],"+"
      jz    @F
      cmp   byte ptr[esi],"-"
      jnz   find_null
      mov   byte ptr[edi+9],80h
   @@:
      inc   esi

find_null:
      mov   eax,esi           ;keep address of most significant digits
   @@:
      cmp   byte ptr[esi],0
      jz    @F
      inc   esi
      jmp   @B
   @@:
      mov   ecx,esi           ;address of terminating 0
      sub   ecx,eax           ;ECX = number of digits in string
      jz    exit_error        ;empty string
      cmp   ecx,19
      mov   edx,ecx
      ja    exit_error        ;too many characters in string
      jnz   pack_bcd
      dec   ecx

;assemble the packed BCD
pack_bcd:
      xor   eax,eax
      dec   esi
      mov   al,[esi]          ;get next character
      sub   al,30h            ;convert ascii to binary
      jc    exit_error        ;non-numerical character
      cmp   al,9
      ja    exit_error        ;non-numerical character
      dec   ecx
      jnz   @F
      stosb
      jmp   use_FPU
   @@:
      mov   ah,al
      dec   esi
      mov   al,[esi]          ;get next character
      sub   al,30h            ;convert ascii to binary
      jc    exit_error        ;non-numerical character
      cmp   al,9
      ja    exit_error        ;non-numerical character
      rol   ah,4
      rol   ax,4
      stosb
      dec   ecx
      jnz   pack_bcd

;check 19th character
      cmp   edx,19
      jnz   use_FPU
      dec   esi
      movzx ecx,byte ptr[esi]
      sub   cl,30h
      jc    exit_error        ;non-numerical character
      cmp   cl,9
      ja    exit_error        ;non-numerical character
      jb    use_FPU
      lea   eax,bcdstring
      cmp   dword ptr[eax+5],22337203h
      ja    exit_error        ;number too large
      jb    use_FPU
      cmp   dword ptr[eax+1],68547758h
      ja    exit_error        ;number too large
      jb    use_FPU
      cmp   byte ptr[eax],7h
      ja    exit_error        ;number too large

use_FPU:
      fclex                   ;clear all exceptions
      ffree st(7)             ;insure the register is free
      fbld  bcdstring         ;load the packed BCD to FPU
      mov   eax,lpqw
      fistp qword ptr[eax]    ;transfer it to memory as a qword
      fstsw ax                ;retrieve exceptions
      shr   al,1              ;check for invalid FPU operation
      jc    exit_error

      cmp   edx,19
      jnz   @F
      mov   eax,1000000000
      mul   eax
      mov   esi,edx
      mul   ecx
      mov   edi,lpqw
      test  byte ptr[edi+7],80h  ;positive or negative
      jnz   negative

      add   [edi],eax
      pushfd
      push  edx
      mov   eax,esi
      mul   ecx
      add   [edi+4],eax
      pop   eax
      popfd
      adc   [edi+4],eax
      jmp   @F

negative:
      sub   [edi],eax
      pushfd
      push  edx
      mov   eax,esi
      mul   ecx
      sub   [edi+4],eax
      pop   eax
      popfd
      sbb   [edi+4],eax

   @@:
      or    al,1              ;no error code
      ret

ascii2qw endp

;end
